{# -*- engine: jinja -*- #}
<div id="summary-{{ link }}" class="summary{% if info.single_page and not info.static_report %}js-enabled-hidden{% endif %}">
  <div>
    <table class="legend">
      <tr>
        <th scope="row">Directory:</th>
        <td>{{info.get_directory()}}</td>
      </tr>
      <tr>
        <th scope="row">Coverage:</th>
        <td class="legend">
          <span class="coverage-low">low: &ge; 0%</span>
          {% if COVERAGE_MED != COVERAGE_HIGH %}
          <span class="coverage-medium">medium: &ge; {{COVERAGE_MED}}%</span>
          {% endif %}
          {% if COVERAGE_HIGH < 100 %}
          <span class="coverage-high">high: &ge; {{COVERAGE_HIGH}}%</span>
          {% else %}
          <span class="coverage-high">high: = 100%</span>
          {% endif %}
        </td>
      </tr>
      {% if (COVERAGE_MED != LINE_COVERAGE_MED) or (COVERAGE_HIGH != LINE_COVERAGE_HIGH) %}
      <tr>
        <th scope="row">Lines:</th>
        <td class="legend">
          <span class="coverage-low">low: &ge; 0%</span>
          {% if LINE_COVERAGE_MED != LINE_COVERAGE_HIGH %}
          <span class="coverage-medium">medium: &ge; {{LINE_COVERAGE_MED}}%</span>
          {% endif %}
          {% if LINE_COVERAGE_HIGH < 100 %}
          <span class="coverage-high">high: &ge; {{LINE_COVERAGE_HIGH}}%</span>
          {% else %}
          <span class="coverage-high">high: = 100%</span>
          {% endif %}
        </td>
      </tr>
      {% endif %}
      {% if (COVERAGE_MED != BRANCH_COVERAGE_MED) or (COVERAGE_HIGH != BRANCH_COVERAGE_HIGH) %}
      <tr>
        <th scope="row">Branches:</th>
        <td class="legend">
          <span class="coverage-low">low: &ge; 0%</span>
          {% if BRANCH_COVERAGE_MED != BRANCH_COVERAGE_HIGH %}
          <span class="coverage-medium">medium: &ge; {{BRANCH_COVERAGE_MED}}%</span>
          {% endif %}
          {% if BRANCH_COVERAGE_HIGH < 100 %}
          <span class="coverage-high">high: &ge; {{BRANCH_COVERAGE_HIGH}}%</span>
          {% else %}
          <span class="coverage-high">high: = 100%</span>
          {% endif %}
        </td>
      </tr>
      {% endif %}
      {% if SHOW_DECISION and decisions.unchecked > 0 %}
      <tr>
        <th scope="row">Warnings:</th>
        <td class="warning_text">{{decisions.unchecked}} unchecked decisions!</td>
      </tr>
      {% endif %}
    </table>
  </div>
  <div>
    <table class="coverage">
      <tr>
        <th></th>
        <th scope="col">Coverage</th>
        <th scope="col">Exec / Excl / Total</th>
      </tr>
      <tr>
        <th scope="row">Lines:</th>
        <td class="{{lines.class}}">{{lines.coverage}}%</td>
        <td class="{{lines.class}}">{{lines.exec}} / {{lines.excluded}} / {{lines.total}}</td>
      </tr>
      <tr>
        <th scope="row">Functions:</th>
        <td class="{{functions.class}}">{{functions.coverage}}%</td>
        <td class="{{functions.class}}">{{functions.exec}} / {{functions.excluded}} / {{functions.total}}</td>
      </tr>
      <tr>
        <th scope="row">Branches:</th>
        <td class="{{branches.class}}">{{branches.coverage}}%</td>
        <td class="{{branches.class}}">{{branches.exec}} / {{branches.excluded}} / {{branches.total}}</td>
      </tr>
      {% if SHOW_CONDITION_COVERAGE %}
      <tr>
        <th scope="row">Conditions:</th>
        <td class="{{conditions.class}}">{{conditions.coverage}}%</td>
        <td class="{{conditions.class}}">{{conditions.exec}} / {{conditions.excluded}} / {{conditions.total}}</td>
      </tr>
      {% endif %}
      {% if SHOW_DECISION %}
      <tr>
        <th scope="row">Decisions:</th>
        <td class="{{decisions.class}}">{{decisions.coverage}}%</td>
        <td class="{{decisions.class}}">{{decisions.exec}} / - / {{decisions.total}}</td>
      </tr>
      {% endif %}
      {% if SHOW_CALLS %}
      <tr>
        <th scope="row">Calls:</th>
        <td class="{{calls.class}}">{{calls.coverage}}%</td>
        <td class="{{calls.class}}">{{calls.exec}} / {{calls.excluded}} / {{calls.total}}</td>
      </tr>
      {% endif %}
    </table>
  </div>
</div>
