@charset "utf-8";

:root {
  font-family: sans-serif;
  --tab_size: {{tab_size}};
}

{% include "style.colors.css" %}


body {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", "Noto Sans", Helvetica, Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji";
  color: var(--color-fg-default);
  background-color: var(--color-canvas-default)
}

h1
{
  text-align: center;
  margin: 0;
  padding-bottom: 10px;
  font-size: 20pt;
  font-weight: bold;
}

hr {
  height: 0;
  margin: 15px 0;
  overflow: hidden;
  background: rgba(0,0,0,0);
  border: 0;
  border-bottom: 1px solid var(--color-border-muted);
}

{% include "style.common.css" %}

{% include "style.summary.css" %}

{% include "style.meter.css" %}

{% include "style.filelist.css" %}

{% include "style.details.css" %}

{% include "style.primer-stub.css" %}
