.color-fg-muted {
  color: var(--color-fg-muted) !important
}

.text-mono {
  font-family: ui-monospace, SFMono-Regular, SF Mono, Menlo, Consolas, Liberation Mono, monospace !important
}
.text-small {
  font-size: 12px !important
}

.Layout-main {
  min-width: 0
}

@media(max-width: 543.98px) {
  .Layout {
    grid-auto-flow: row;
    grid-template-columns: 1fr !important
  }

  .Layout .Layout-main {
    grid-column: 1 !important;
    width: 100% !important
  }
}

.Layout-main .Layout-main-centered-lg {
  margin-left: auto;
  margin-right: auto
}

.Box {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", "Noto Sans", Helvetica, Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji";
  font-size: var(--body-font-size, 14px);
  line-height: 1.5;
  color: var(--color-fg-default);
  background-color: var(--color-canvas-default);
  border-color: var(--color-border-default);
  border-radius: var(--borderRadius-medium, 6px);
  border-style: solid;
  border-width: var(--borderWidth-thin, 1px)
}

.Box-header {
  background-color: var(--color-canvas-subtle);
  border-color: var(--color-border-default);
  border-style: solid;
  border-top-left-radius: 6px;
  border-top-right-radius: 6px;
  border-width: 1px;
  margin: -1px -1px 0;
  padding: 16px;
}

.Box-row {
  border-top: var(--borderWidth-thin, 1px) solid var(--color-border-muted);
  list-style-type: none;
  margin-top: calc(var(--borderWidth-thin, 1px)*-1);
  padding: var(--stack-padding-normal, 16px)
}

.Box-row:first-of-type {
  border-top-left-radius: var(--borderRadius-medium, 6px);
  border-top-right-radius: var(--borderRadius-medium, 6px)
}

.Box-row:last-of-type {
  border-bottom-left-radius: var(--borderRadius-medium, 6px);
  border-bottom-right-radius: var(--borderRadius-medium, 6px)
}

.Box-row--focus-gray:hover {
  background-color: var(--color-canvas-subtle) !important;
}

.filelist .Box-row:nth-child(2n+3){
  background-color: var(--color-accent-subtle);
}

.functionslist .Box-row:nth-child(2n+3){
  background-color: var(--color-accent-subtle);
}

.d-flex {
  display: flex !important
}

.flex-auto {
  flex: auto !important
}

.flex-justify-end {
  justify-content: flex-end !important
}

.flex-justify-center {
  justify-content: center !important
}

.flex-wrap {
  flex-wrap: wrap !important
}

.flex-items-baseline {
  align-items: baseline !important
}

.flex-space-between {
  justify-content: space-between !important
}

.min-width-0 {
  min-width: 0 !important
}

.m-3 {
  margin: var(--base-size-16, 16px) !important
}

.py-2 {
  padding-top: var(--base-size-8, 8px) !important;
  padding-bottom: var(--base-size-8, 8px) !important;
}

.col-2,
.col-md-2
{
  width: 12.5%
}

.css-truncate.css-truncate-target {
  overflow: hidden;
  text-overflow: ellipsis;
  white-space: nowrap;
  display: inline-block;
  max-width: 125px;
  vertical-align: top
}

.d-block {
  display: block !important
}

.width-fit {
  max-width: 100% !important
}

.text-center {
  text-align: center !important
}
