/*** Summary formats ***/

.summary
{
  display: flex;
  flex-flow: row wrap;
  max-width: 100%;
  justify-content: flex-start;
}

.summary > table
{
  flex: 1 0 7em;
  border: 0;
}

.summary > :last-child {
  margin-left: auto;
}

table.legend
{
  color: black;
  display: flex;
  flex-flow: row wrap;
  justify-content: flex-start;
  margin-left: 16px;
}

table.legend th[scope=row]
{
  font-weight: normal;
  text-align: right;
  white-space: nowrap;
}

table.legend td
{
  color: blue;
  text-align: left;
  white-space: nowrap;
  padding-left: 5px;
}

table.legend td.legend
{
  color: black;
  font-size: 80%;
}

table.legend td.warning_text
{
  color: var(--warning_color);
}

table.coverage
{
  border-collapse: collapse;
  margin-right: 16px;
}

table.coverage td,
table.coverage th
{
  text-align: right;
  color: black;
  font-weight: normal;
  white-space: nowrap;
  padding-top: 3px;
  padding-bottom: 3px;
  padding-left: 5px;
  padding-right: 4px;
}

table.coverage td
{
  background-color: LightSteelBlue;
}

table.coverage th[scope=row]
{
  color: black;
  font-weight: normal;
  white-space: nowrap;
}

table.coverage th[scope=col]
{
  color: blue;
  font-weight: normal;
  white-space: nowrap;
}

table.legend span
{
  margin-right: 4px;
  padding: 2px;
}

table.legend span.coverage-unknown,
table.legend span.coverage-none,
table.legend span.coverage-low,
table.legend span.coverage-medium,
table.legend span.coverage-high
{
  padding-left: 0.5em;
  padding-right: 0.5em;
}

table.legend span.coverage-unknown,
table.coverage td.coverage-unknown
{
  background-color: var(--unknown_color) !important;
}

table.legend span.coverage-none,
table.legend span.coverage-low,
table.coverage td.coverage-none,
table.coverage td.coverage-low
{
  background-color: var(--low_color) !important;
}

table.legend span.coverage-medium,
table.coverage td.coverage-medium
{
  background-color: var(--medium_color) !important;
}

table.legend span.coverage-high,
table.coverage td.coverage-high
{
  background-color: var(--high_color) !important;
}

/*** End of Summary formats ***/
